"""Pipeline class definitions"""
from collections import defaultdict
from datetime import datetime
import logging
from typing import Any, Callable, Iterable, Optional, Tuple, Union

from joblib import delayed, Parallel
from numpy import iterable
import pandas as pd
from tqdm import tqdm

from wildebeest.constants import PathOrStr


RUN_REPORT_COLS = ['outpath', 'skipped', 'error', 'time_finished']


class Pipeline:
    """
    Class for defining file processing pipelines.

    Attributes
    ----------
    load_func
        Callable that takes a string or `Path` object as single
        positional argument, reads from the corresponding location, and
        returns some representation of its contents.
    write_func
        Callable that takes the output of the last element of `ops` (or
        the output of `load_func` if `ops` is `None` or empty) and a
        string or `Path` object and writes the former to the location
        specified by the latter.
    ops
        Iterable of callables each of which takes a single positional
        argument. The first element of `ops` must accept the output of
        `load_func`, and each subsequent element must accept the output
        of the immediately preceding element. It is recommended that
        every element of `ops` take and return one common data structure
        (e.g. NumPy arrays for image data) so that those elements can
        be recombined easily.
    run_report_
    """

    def __init__(
        self,
        load_func: Callable[[PathOrStr], Any],
        write_func: Callable[[Any, PathOrStr], None],
        ops: Optional[
            Union[Callable[[Any], Any], Iterable[Callable[[Any], Any]]]
        ] = None,
    ) -> None:
        self.load_func = load_func
        self.write_func = write_func
        self.ops = self._process_ops(ops)
        self._run_report_ = None
        self._log_dict = defaultdict(dict)

    def _process_ops(self, ops):
        if iterable(ops):
            return ops
        elif callable(ops):
            return [ops]
        elif ops is None:
            return []
        else:
            raise TypeError('ops must be callable, an iterable of callables, or `None`')

    @property
    def run_report_(self):
        """
        Pandas DataFrame of information about the most recent run.

        Stores input path in the index, output path as "outpath",
        Boolean indicating whether the file was skipped as "skipped",
        the repr of an exception object that was handled during
        processing if any as "error" (`np.nan` if no exception was
        handled), and a timestamp indicating when processing completed
        as "time_finished".

        May include additional custom fields in a
        `CustomReportingPipeline`.

        Raises
        ------
        AttributeError
            If no run report is available because pipeline has not been
            run.
        """
        if self._run_report_ is None:
            raise AttributeError(
                'Pipeline has not been run, so there is no run report.'
            )
        return self._run_report_

    def __call__(
        self,
        inpaths: Iterable[PathOrStr],
        path_func: Callable[[PathOrStr], PathOrStr],
        n_jobs: int,
        skip_func: Optional[Callable[[PathOrStr, Any], bool]] = None,
        exceptions_to_catch: Optional[Union[Exception, Tuple[Exception]]] = Exception,
    ) -> pd.DataFrame:
        """
        Run the pipeline.

        Across `n_jobs` threads, for each path in `inpaths`, if
        `skip_func(path, path_func(path))` is `True`, skip that path.
        Otherwise, use `load_func` to get the resource from that path,
        pipe its output through `ops`, and write out the result with
        `write_func`.

        Parameters
        ----------
        inpaths
            Iterable of string or Path objects pointing to resources to
            be processed and written out.
        path_func
            Function that takes each input path and returns the desired
            corresponding output path.
        n_jobs
            Number of threads to use.
        skip_func
            Callable that takes an input path and a prospective output
            path and returns a Boolean indicating whether or not that
            file should be skipped; for instance, use
            `lambda inpath, outpath: Path(outpath).is_file()` to avoid
            overwriting existing files.
        exceptions_to_catch
            Tuple of exception types to catch. An exception of one of
            these types will be logged with logging level ERROR and
            added to the run report, but the pipeline will continue to
            execute.

        Note
        ----
        Stores a run report in `self.run_report_`
        """
        Parallel(n_jobs=n_jobs, prefer='threads')(
            delayed(self._pipeline_func)(
                path, path_func, skip_func, exceptions_to_catch
            )
            for path in tqdm(inpaths)
        )

        logging.info('Processing finished. Creating run report.')
        self._run_report_ = pd.DataFrame.from_dict(self._log_dict, orient='index')
        self._log_dict = defaultdict(dict)
        # Default ns precision is overkill for most applications and
        # gives an error when writing to parquet.
        self._run_report_.loc[:, 'time_finished'] = self._run_report_.loc[
            :, 'time_finished'
        ].astype('datetime64[ms]')
        self._run_report_ = self._run_report_.reindex(
            columns=(
                RUN_REPORT_COLS
                + [col for col in self._run_report_ if col not in RUN_REPORT_COLS]
            )
        )
        self._run_report_.loc[:, "skipped"] = self._run_report_.loc[
            :, "skipped"
        ].fillna(False)

    def _pipeline_func(
        self,
        inpath: PathOrStr,
        outpath_func: PathOrStr,
        skip_func: Callable,
        exceptions_to_catch: Optional[Union[Exception, Tuple[Exception]]] = Exception,
    ) -> None:
        """
        Process one file

        Use `self.load_func` to load the file at `inpath` into memory,
        pipe the result through the functions in `self.ops`, and use
        `self.write_func` to write it to `outpath_func(inpath)`.

        If `skip_existing` is `True`, check up front whether
        `self.check_existing_func(outpath_func(inpath))` exists. If it
        does, skip the file.

        Catch `exceptions_to_catch` if they arise during file
        processing.

        Parameters
        ----------
        inpath
            Input path
        outpath_func
            Function that takes input path and returns output path
        skip_existing
            Whether or not to skip processing if output path is occupied
        exceptions_to_catch
            Exception types to catch

        Note
        ----
        Records results in a dict within `self._log_dict[inpath]` with
        the following keys

        outpath
            output path
        skipped_existing
            0/1 indicating whether existing file was skipped
        time_finished
            Timestamp indicating when processing finished
        """

        class _DummyException(Exception):
            pass

        if exceptions_to_catch is None:
            exceptions_to_catch = _DummyException

        self._log_dict[inpath]['outpath'] = outpath_func(inpath)
        if skip_func is not None and skip_func(
            inpath, self._log_dict[inpath]['outpath']
        ):
            self._log_dict[inpath]['skipped'] = True
            logging.debug(
                f'Skipping {inpath} because there is already a file at corresponding '
                f'output path {self._log_dict[inpath]["outpath"]}'
            )
            self._log_dict[inpath]['time_finished'] = datetime.now()
        else:
            try:
                self._run_pipeline_func(inpath, self._log_dict[inpath]['outpath'])
            except exceptions_to_catch as e:
                self._log_dict[inpath]['error'] = repr(e)
                logging.error(f'{type(e)} exception on {inpath}: {e}')
            finally:
                self._log_dict[inpath]['time_finished'] = datetime.now()

    def _run_pipeline_func(self, inpath, outpath):
        stage = self.load_func(inpath)
        for op in self.ops:
            stage = op(stage)
        self.write_func(stage, outpath)


class CustomReportingPipeline(Pipeline):
    """
    Class for defining file processing pipelines with custom run
    recording.

    Differences from Pipeline parent class:

        `load_func`, each element of `ops`, and `write_func` must each
        accept the string or `Path` object indicating the input item's
        location as an additional positional argument.

        Each element of `ops` and `write_func` must each accept a
        `defaultdict(dict)` object as an additional positional argument.
        Functions defined in Wildebeest call this item `log_dict`.

    Inside those functions, adding items to `log_dict[inpath]` causes
    them to be added to the "run record" DataFrame that the pipeline
    returns.
    """

    def _run_pipeline_func(self, inpath, outpath):
        stage = self.load_func(inpath, log_dict=self._log_dict)
        for op in self.ops:
            stage = op(stage, inpath=inpath, log_dict=self._log_dict)
        self.write_func(stage, outpath, inpath=inpath, log_dict=self._log_dict)



import pytest

from tests.conftest import SAMPLE_DATA_DIR
from wildebeest.load_funcs.image import load_image_from_disk


@pytest.fixture
def sample_image_square_rgb():
    filename = 'wildebeest_rgb.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_tall_grayscale():
    filename = 'wildebeest_gray.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgba():
    filename = 'wildebeest_rgba.png'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgba_flipped_horiz():
    filename = 'wildebeest_rgba_flipped_horiz.png'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgba_flipped_vert():
    filename = 'wildebeest_rgba_flipped_vert.png'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgb_flipped_horiz():
    filename = 'wildebeest_rgb_flipped_horiz.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgb_flipped_vert():
    filename = 'wildebeest_rgb_flipped_vert.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_grayscale_flipped_horiz():
    filename = 'wildebeest_gray_flipped_horiz.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_grayscale_flipped_vert():
    filename = 'wildebeest_gray_flipped_vert.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_tall_grayscale_rotated_90():
    filename = 'wildebeest_gray_rotated_90.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgb_rotated_90():
    filename = 'wildebeest_rgb_rotated_90.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgba_rotated_90():
    filename = 'wildebeest_rgba_rotated_90.png'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_tall_grayscale_rotated_180():
    filename = 'wildebeest_gray_rotated_180.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgb_rotated_180():
    filename = 'wildebeest_rgb_rotated_180.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgba_rotated_180():
    filename = 'wildebeest_rgba_rotated_180.png'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_tall_grayscale_rotated_270():
    filename = 'wildebeest_gray_rotated_270.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgb_rotated_270():
    filename = 'wildebeest_rgb_rotated_270.jpg'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)


@pytest.fixture
def sample_image_square_rgba_rotated_270():
    filename = 'wildebeest_rgba_rotated_270.png'
    path = SAMPLE_DATA_DIR / filename
    return load_image_from_disk(path)



from functools import partial
from pathlib import Path
import sys

from wildebeest.path_funcs import join_outdir_filename_extension

TEST_DIR = Path(__file__).parent
SAMPLE_DATA_DIR = Path(TEST_DIR) / 'sample_data'
TEMP_DATA_DIR = SAMPLE_DATA_DIR / 'tmp'
IMAGE_FILENAMES = ['2RsJ8EQ', '2TqoToT', '2VocS58', '2scKPIp', '2TsO6Pc', '2SCv0q7']
IMAGE_URLS = [f'https://bit.ly/{filename}' for filename in IMAGE_FILENAMES]

from tests.fixtures import *

sys.path.append(TEST_DIR / 'fixtures')

keep_filename_save_png_in_tempdir = partial(
    join_outdir_filename_extension, outdir=TEMP_DATA_DIR, extension='.png'
)


def delete_file_if_exists(path: Path):
    if path.is_file():
        path.unlink()



from shutil import rmtree

import pytest

from tests.conftest import TEMP_DATA_DIR


@pytest.fixture(scope='session')
def generate_file_tree():
    if TEMP_DATA_DIR.exists():
        rmtree(str(TEMP_DATA_DIR))
    level0_dirs = ['tmp00', 'tmp01']
    level0_files = ['fake00.txt', 'fake01.png', 'fake02.JPG']
    for temp_dir in level0_dirs:
        dirpath = TEMP_DATA_DIR / temp_dir
        dirpath.mkdir(parents=True)
    for filename in level0_files:
        filepath = TEMP_DATA_DIR / filename
        filepath.touch()
    level2_dirpath = TEMP_DATA_DIR / level0_dirs[0] / 'temp10' / 'temp20'
    level2_dirpath.mkdir(parents=True)
    level3_files = ['fake30.pdf', 'fake31.BMP', 'fake32.txt']
    for filename in level3_files:
        filepath = level2_dirpath / filename
        filepath.touch()
    yield
    rmtree(str(TEMP_DATA_DIR))



import operator

import cv2 as cv
import numpy as np
import pytest

from wildebeest.ops.image import (
    centercrop,
    flip_horiz,
    flip_vert,
    resize,
    rotate_180,
    rotate_270,
    rotate_90,
    trim_padding,
)


class TestResize:
    def test_resize_validation(self, sample_image_square_rgb):
        with pytest.raises(ValueError):
            resize(sample_image_square_rgb, shape=(10, 10), min_dim=5)

    def test_resize_with_shape(self, sample_image_square_rgb):
        expected_shape = (5, 8)
        image = resize(sample_image_square_rgb, shape=expected_shape)
        actual_shape = image.shape[:2]
        assert actual_shape == expected_shape

    def test_resize_square_with_min_dim(self, sample_image_square_rgb):
        image = resize(sample_image_square_rgb, min_dim=50)
        actual_shape = image.shape[:2]
        expected_shape = (50, 50)
        assert actual_shape == expected_shape

    def test_resize_tall_with_min_dim(self, sample_image_tall_grayscale):
        image = resize(sample_image_tall_grayscale, min_dim=50)
        actual_shape = image.shape
        expected_shape = (68, 50)
        assert actual_shape == expected_shape

    def test_resize_wide_with_min_dim(self, sample_image_tall_grayscale):
        sample_image_wide = sample_image_tall_grayscale.transpose()
        image = resize(sample_image_wide, min_dim=50)
        actual_shape = image.shape
        expected_shape = (50, 68)
        assert actual_shape == expected_shape

    def test_resize_accepts_custom_reporting_args(self, sample_image_square_rgb):
        resize(sample_image_square_rgb, shape=(5, 8), inpath='fake', log_dict={})


class TestCenterCrop:
    def test_centercrop_square_rgb(self, sample_image_square_rgb):
        expected_shape = (16, 16, 3)
        actual_shape = centercrop(sample_image_square_rgb, reduction_factor=0.5).shape
        assert expected_shape == actual_shape

    def test_centercrop_tall_greyscale(self, sample_image_tall_grayscale):
        expected_shape = (17, 12)
        actual_shape = centercrop(
            sample_image_tall_grayscale, reduction_factor=0.5
        ).shape
        assert expected_shape == actual_shape

    def test_centercrop_square_rgba(self, sample_image_square_rgba):
        expected_shape = (16, 16, 4)
        actual_shape = centercrop(sample_image_square_rgba, reduction_factor=0.5).shape
        assert expected_shape == actual_shape

    def test_centercrop_reduction_factor(self, sample_image_square_rgb):
        expected_shape = (19, 19, 3)
        actual_shape = centercrop(sample_image_square_rgb, reduction_factor=0.6).shape
        assert expected_shape == actual_shape

    def test_centercrop_accepts_custom_reporting_args(self, sample_image_square_rgb):
        centercrop(
            sample_image_square_rgb, reduction_factor=0.5, inpath='fake', log_dict={}
        )


class TestTrimPadding:
    def test_trim_padding_no_padding(self, sample_image_square_rgb):
        image = trim_padding(
            sample_image_square_rgb, thresh=0.95, comparison_op=operator.gt
        )
        np.testing.assert_almost_equal(image, sample_image_square_rgb)

    def test_trim_padding_rgb(self, sample_image_square_rgb):
        im_padded = cv.copyMakeBorder(
            src=sample_image_square_rgb,
            top=np.random.randint(1, 100),
            bottom=np.random.randint(1, 100),
            left=np.random.randint(1, 100),
            right=np.random.randint(1, 100),
            borderType=cv.BORDER_CONSTANT,
            value=[251, 252, 253],
        )
        actual = trim_padding(im_padded, thresh=0.95, comparison_op=operator.gt)
        np.testing.assert_almost_equal(actual, sample_image_square_rgb)

    def test_trim_padding_rgba(self, sample_image_square_rgba):
        im_padded = cv.copyMakeBorder(
            src=sample_image_square_rgba,
            top=np.random.randint(1, 100),
            bottom=np.random.randint(1, 100),
            left=np.random.randint(1, 100),
            right=np.random.randint(1, 100),
            borderType=cv.BORDER_CONSTANT,
            value=[251, 252, 253],
        )
        actual = trim_padding(im_padded, thresh=0.95, comparison_op=operator.gt)
        np.testing.assert_almost_equal(actual, sample_image_square_rgba)

    def test_trim_padding_grayscale(self, sample_image_tall_grayscale):
        im_padded = cv.copyMakeBorder(
            src=sample_image_tall_grayscale,
            top=np.random.randint(1, 100),
            bottom=np.random.randint(1, 100),
            left=np.random.randint(1, 100),
            right=np.random.randint(1, 100),
            borderType=cv.BORDER_CONSTANT,
            value=251,
        )
        actual = trim_padding(im_padded, thresh=0.95, comparison_op=operator.gt)
        np.testing.assert_almost_equal(actual, sample_image_tall_grayscale)

    def test_trim_black_padding_rgb(self, sample_image_square_rgb):
        im_padded = cv.copyMakeBorder(
            src=sample_image_square_rgb,
            top=np.random.randint(1, 100),
            bottom=np.random.randint(1, 100),
            left=np.random.randint(1, 100),
            right=np.random.randint(1, 100),
            borderType=cv.BORDER_CONSTANT,
            value=3,
        )
        actual = trim_padding(im_padded, thresh=0.05, comparison_op=operator.lt)
        np.testing.assert_almost_equal(actual, sample_image_square_rgb)


class TestFlipHoriz:
    def test_flip_horiz_rgba(
        self, sample_image_square_rgba, sample_image_square_rgba_flipped_horiz
    ):
        image = flip_horiz(sample_image_square_rgba)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgba_flipped_horiz, decimal=-1
        )

    def test_flip_horiz_rgb(
        self, sample_image_square_rgb, sample_image_square_rgb_flipped_horiz
    ):
        image = flip_horiz(sample_image_square_rgb)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgb_flipped_horiz, decimal=-1
        )

    def test_flip_horiz_grayscale(
        self, sample_image_tall_grayscale, sample_image_grayscale_flipped_horiz
    ):
        image = flip_horiz(sample_image_tall_grayscale)
        np.testing.assert_almost_equal(
            image, sample_image_grayscale_flipped_horiz, decimal=-1
        )


class TestFlipVert:
    def test_flip_vert_rgba(
        self, sample_image_square_rgba, sample_image_square_rgba_flipped_vert
    ):
        image = flip_vert(sample_image_square_rgba)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgba_flipped_vert, decimal=-1
        )

    def test_flip_vert_rgb(
        self, sample_image_square_rgb, sample_image_square_rgb_flipped_vert
    ):
        image = flip_vert(sample_image_square_rgb)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgb_flipped_vert, decimal=-1
        )

    def test_flip_vert_grayscale(
        self, sample_image_tall_grayscale, sample_image_grayscale_flipped_vert
    ):
        image = flip_vert(sample_image_tall_grayscale)
        np.testing.assert_almost_equal(
            image, sample_image_grayscale_flipped_vert, decimal=-1
        )


class TestRotate_90:
    def test_rotate_rgba_90(
        self, sample_image_square_rgba, sample_image_square_rgba_rotated_90
    ):
        image = rotate_90(sample_image_square_rgba)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgba_rotated_90, decimal=-1
        )

    def test_rotate_rgb_90(
        self, sample_image_square_rgb, sample_image_square_rgb_rotated_90
    ):
        image = rotate_90(sample_image_square_rgb)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgb_rotated_90, decimal=-1
        )

    def test_rotate_grayscale_90(
        self, sample_image_tall_grayscale, sample_image_tall_grayscale_rotated_90
    ):
        image = rotate_90(sample_image_tall_grayscale)
        np.testing.assert_almost_equal(
            image, sample_image_tall_grayscale_rotated_90, decimal=-2
        )


class TestRotate_180:
    def test_rotate_rgba_180(
        self, sample_image_square_rgba, sample_image_square_rgba_rotated_180
    ):
        image = rotate_180(sample_image_square_rgba)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgba_rotated_180, decimal=-1
        )

    def test_rotate_rgb_180(
        self, sample_image_square_rgb, sample_image_square_rgb_rotated_180
    ):
        image = rotate_180(sample_image_square_rgb)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgb_rotated_180, decimal=-1
        )

    def test_rotate_grayscale_180(
        self, sample_image_tall_grayscale, sample_image_tall_grayscale_rotated_180
    ):
        image = rotate_180(sample_image_tall_grayscale)
        np.testing.assert_almost_equal(
            image, sample_image_tall_grayscale_rotated_180, decimal=-2
        )


class TestRotate_270:
    def test_rotate_rgba_270(
        self, sample_image_square_rgba, sample_image_square_rgba_rotated_270
    ):
        image = rotate_270(sample_image_square_rgba)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgba_rotated_270, decimal=-1
        )

    def test_rotate_rgb_270(
        self, sample_image_square_rgb, sample_image_square_rgb_rotated_270
    ):
        image = rotate_270(sample_image_square_rgb)
        np.testing.assert_almost_equal(
            image, sample_image_square_rgb_rotated_270, decimal=-2
        )

    def test_rotate_grayscale_270(
        self, sample_image_tall_grayscale, sample_image_tall_grayscale_rotated_270
    ):
        image = rotate_270(sample_image_tall_grayscale)
        np.testing.assert_almost_equal(
            image, sample_image_tall_grayscale_rotated_270, decimal=-2
        )



"""Functions that take an image and return a transformed image"""
from typing import Callable, Optional, Tuple

import cv2 as cv
import numpy as np


def resize(
    image: np.array,
    shape: Optional[Tuple[int, int]] = None,
    min_dim: Optional[int] = None,
    **kwargs,
) -> np.array:
    """
    Resize input image

    `shape` or `min_dim` needs to be specified with `partial` before
    this function can be used in a Wildebeest pipeline.

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    image
        NumPy array with two spatial dimensions and optionally an
        additional channel dimension
    shape
        Desired output shape in pixels in the form (height, width)
    min_dim
        Desired minimum spatial dimension in pixels; image will be
        resized so that it has this length along its smaller spatial
        dimension while preseving aspect ratio as closely as possible.
        Exactly one of `shape` and `min_dim` must be `None`.
    """
    _validate_resize_inputs(shape, min_dim)
    if min_dim is not None:
        shape = _find_min_dim_shape(image, min_dim)
    return cv.resize(image, dsize=shape[::-1])


def _validate_resize_inputs(shape, min_dim) -> None:
    if (shape is None) + (min_dim is None) != 1:
        raise ValueError('Exactly one of `shape` and `min_dim` must be None')


def centercrop(image: np.array, reduction_factor: float, **kwargs) -> np.array:
    """
    Crop the center out of an image

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    image
        Numpy array of an image. Function will handle 2D greyscale
        images, RGB, and RGBA image arrays
    reduction_factor
        scale of center cropped box, 1.0 would be the full image
        value of .4 means a box of .4*width and .4*height
    """
    height, width, *channels = image.shape

    w_scale = width * reduction_factor
    h_scale = height * reduction_factor

    left = int((width - w_scale) // 2)
    top = int((height - h_scale) // 2)
    right = int((width + w_scale) // 2)
    bottom = int((height + h_scale) // 2)

    return image[top:bottom, left:right]


def trim_padding(
    image: np.array, comparison_op: Callable, thresh: int, **kwargs
) -> np.array:
    """
    Remove padding from an image

    Remove rows and columns on the edges of the input image where the
    brightness on a scale of 0 to 1 satisfies `comparison_op` with
    respect to `thresh`. Brightness is evaluated by converting to
    grayscale and normalizing if necessary. For instance, using
    `thresh=.95` and `comparison_op=operator.gt` will result in removing
    near-white padding, while using using `thresh=.05` and
    `comparison_op=operator.lt` will remove near-black padding.

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Assumes:

        Image is grayscale, RGB, or RGBA.

        Pixel values are scaled between either 0 and 1 or 0 and 255. If
        image is scaled between 0 and 255, then some pixel has a value
        greater than 1.

    Parameters
    ----------
    image
        Numpy array of an image.
    comparison_op
        How to compare pixel values to `thresh`
    thresh
        Value to compare pixel values against
    """
    im_gray = convert_to_grayscale(image)
    im_gray = normalize_pixel_values(im_gray)
    keep = ~comparison_op(im_gray, thresh)
    x, y, w, h = cv.boundingRect(cv.findNonZero(keep.astype(int)))
    return image[y : y + h, x : x + w]


def normalize_pixel_values(image: np.array) -> np.array:
    """
    Normalize image so that pixel values are between 0 and 1

    Assumes pixel values are scaled between either 0 and 1 or 0 and 255.
    """
    if image.max() > 1:
        return image / 255
    else:
        return image


def convert_to_grayscale(image: np.array) -> np.array:
    """
    Convert image to grayscale.

    Assumes image is grayscale, RGB, or RGBA.
    """
    grayscale = image.ndim == 2
    if grayscale:
        im_gray = image
    else:
        rgba = image.shape[2] == 4
        if rgba:
            im_gray = cv.cvtColor(image, cv.COLOR_RGBA2GRAY)
        else:
            im_gray = cv.cvtColor(image, cv.COLOR_RGB2GRAY)
    return im_gray


def _find_min_dim_shape(image, min_dim):
    in_height, in_width = image.shape[:2]
    aspect_ratio = in_width / in_height
    format = 'tall' if aspect_ratio < 1 else 'wide'
    if format == 'tall':
        out_width = min_dim
        out_height = round(out_width / aspect_ratio, 1)
    else:
        out_height = min_dim
        out_width = round(out_height * aspect_ratio, 1)
    return (int(out_height), int(out_width))


def flip_horiz(image: np.array) -> np.array:
    """Flip an image horizontally"""
    return cv.flip(image, flipCode=1)


def flip_vert(image: np.array) -> np.array:
    """Flip an image vertically"""
    return cv.flip(image, flipCode=0)


def rotate_90(image: np.array) -> np.array:
    """
    Rotate an image 90 degrees counterclockwise

    This function takes an image as numpy array and
    and outputs the image rotated 90 degrees counterclockwise.

    Assumes that the image is going to be rotated around center, and size of image
    will remain unchanged.

    This function takes numpy array of an image. Function will handle 2D greyscale
    images, RGB, and RGBA image arrays.
    """
    return cv.rotate(image, cv.ROTATE_90_COUNTERCLOCKWISE)


def rotate_180(image: np.array) -> np.array:
    """
    Rotate an image 180 degrees

    This function takes an image as numpy array and
    and outputs the image rotated 180 degrees.

    Assumes that the image is going to be rotated around center, and size of image
    will remain unchanged.

    This function takes numpy array of an image. Function will handle 2D greyscale
    images, RGB, and RGBA image arrays.
    """
    return cv.rotate(image, cv.ROTATE_180)


def rotate_270(image: np.array) -> np.array:
    """
    Rotate an image 270 degrees counterclockwise

    This function takes an image as numpy array and
    and outputs the image rotated 270 degrees counterclockwise.

    Assumes that the image is going to be rotated around center, and size of image
    will remain unchanged.

    This function takes numpy array of an image. Function will handle 2D greyscale
    images, RGB, and RGBA image arrays.
    """
    return cv.rotate(image, cv.ROTATE_90_CLOCKWISE)



"""Miscellaneous utilities for images"""
from functools import partial
import mimetypes

from wildebeest.util import find_files_with_extensions

find_image_files = partial(
    find_files_with_extensions,
    extensions=[k for k, v in mimetypes.types_map.items() if v.startswith("image/")],
)
"""Find all image files in a directory"""



"""Miscellaneous utilities"""
import os
from pathlib import Path
from typing import Iterable, List

from wildebeest.constants import PathOrStr


def find_files_with_extensions(
    search_dir: PathOrStr, extensions: Iterable[str]
) -> List[Path]:
    """
    Find files with one of the specified extensions.

    Extension matching is case-insensitive.

    Parameters
    ----------
    search_dir
        Directory to search
    img_extensions
        Extensions to search for. The initial "." can be included or
        not.

    Returns
    -------
    list
        List of `Path` objects specifying locations of all files
        recursively within `search_dir` that have one of the extensions
        in `extensions`.
    """
    extensions = [item if item.startswith('.') else '.' + item for item in extensions]
    paths = []
    for dirpath, _, filenames in os.walk(str(search_dir)):
        for fn in filenames:
            if Path(fn).suffix.lower() in extensions:
                paths.append(Path(dirpath) / fn)
    return paths



"""Functions that take an image and write it out"""
import os
from pathlib import Path
import uuid

import cv2 as cv
import numpy as np

from wildebeest.constants import PathOrStr


def write_image(image: np.array, path: PathOrStr, **kwargs) -> None:
    """
    Write image to specified path.

    Create output directory if it does not exist.

    Write to a temporary in a directory ".tmp" inside the output
    directory and then rename the file so that we don't create a partial
    image file if write process is interrupted. ".tmp" directory is not
    deleted, but temporary files are deleted even if there is an
    exception during writing or renaming.

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    image
        Image as a NumPy array
    path
        Desired output path
    """
    num_channels = 1 if len(image.shape) == 2 else image.shape[2]
    if num_channels >= 3:
        # OpenCV wants to write BGR, so reverse order of first three
        # channels
        image[:, :, :3] = image[:, :, 2::-1]

    path = Path(path)
    outdir = path.parent

    tmp_dir = outdir / '.tmp'
    tmp_dir.mkdir(parents=True, exist_ok=True)
    tmp_path = tmp_dir / (str(uuid.uuid4()) + path.suffix)
    try:
        cv.imwrite(str(tmp_path), image)
        assert (
            tmp_path.exists()
        ), f"Attempt to write image temporarily to {tmp_path} failed."
        os.rename(tmp_path, path)
    finally:
        if tmp_path.exists():
            tmp_path.unlink()



"""Image-processing pipelines"""
from wildebeest import Pipeline
from wildebeest.load_funcs.image import load_image_from_url
from wildebeest.write_funcs.image import write_image


download_image_pipeline = Pipeline(
    load_func=load_image_from_url, ops=[], write_func=write_image
)
"""Basic pipeline for downloading images"""
